



pip check
IF %ERRORLEVEL% NEQ 0 exit /B 1
cd tests
IF %ERRORLEVEL% NEQ 0 exit /B 1
set SKIP_OMP=true
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest -vv test_imagecodecs.py -k "not (test_cms_profile or test_jpeg_rgb_mode)"
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
